"""
Functions for EM computation
Author: Sergey Bobkov
"""
import numpy as np
from numba import njit, prange


@njit(parallel=True, fastmath=True)
def extract_all_model(model, q_vals, t_vals, rotations, out_data):
    """Extract model for all classes in all rotations

    Keyword arguments:
    model -- 3D model array (num_class, y, x)
    q_vals -- radius values for extraction
    t_vals -- angle values for extraction
    rotations -- array with values of angles for rotations
    out_data -- result will be written here
    """
    num_class = len(model)
    num_rot = len(rotations)
    for i in prange(num_class*num_rot):
        k = i // num_rot
        j = i % num_rot
        rot = rotations[j]
        extract_model(model[k], q_vals, t_vals + rot, out_data[k, j])


@njit(fastmath=True)
def extract_model(model, q_vals, t_vals, out_data):
    """Extract model for particular class in particular rotation

    Keyword arguments:
    model -- 2D model array (y, x)
    q_vals -- radius values for extraction
    t_vals -- angle values for extraction
    out_data -- result will be written here
    """
    assert len(q_vals) == len(t_vals)

    center = (model.shape[0]-1) // 2
    out_data[:] = 0

    for i in range(len(q_vals)):
        x_val = q_vals[i]*np.cos(t_vals[i]) + center
        y_val = q_vals[i]*np.sin(t_vals[i]) + center

        x_fl = int(np.floor(x_val))
        y_fl = int(np.floor(y_val))

        x_frac = x_val - x_fl
        y_frac = y_val - y_fl

        res = 0

        if test_valid(y_fl, model.shape[0]) and test_valid(x_fl, model.shape[1]):
            res += (1 - x_frac)*(1 - y_frac)*model[y_fl, x_fl]

        if test_valid(y_fl, model.shape[0]) and test_valid(x_fl + 1, model.shape[1]):
            res += (x_frac)*(1 - y_frac)*model[y_fl, x_fl + 1]

        if test_valid(y_fl + 1, model.shape[0]) and test_valid(x_fl, model.shape[1]):
            res += (1 - x_frac)*(y_frac)*model[y_fl + 1, x_fl]

        if test_valid(y_fl + 1, model.shape[0]) and test_valid(x_fl + 1, model.shape[1]):
            res += (x_frac)*(y_frac)*model[y_fl + 1, x_fl + 1]

        out_data[i] = res


@njit
def test_valid(value, max_val):
    """Test if 0 <= value < max_val"""
    if 0 <= value < max_val:
        return True
    return False


@njit(parallel=True, fastmath=True) # fastmath=True cause dead stuck with MPI
def compute_all_loglh(data_vals, data_idx, data_scale, model_slice, out_loglh):
    """Compute log-likelihood for all frames against all classes in all orientations

    Keyword arguments:
    data_vals -- values of data intensity
    data_idx -- indexes for data values against model_slice values
    data_scale -- scale of individual frames in data
    model_slice -- slice of model computed at all requred rotations
    num_class -- number of classes (number of models)
    num_rot -- number of rotations
    out_loglh -- result will be written here
    """
    num_class, num_rot = model_slice.shape[:2]
    for i in prange(num_class*num_rot):
        k = i // num_rot
        j = i % num_rot
        compute_loglh(data_vals, data_idx, data_scale,
                      model_slice[k, j], out_loglh[k, j])


@njit(fastmath=True) # fastmath=True cause dead stuck with MPI
def compute_loglh(data_vals, data_idx, data_scale,
                  model_slice, out_loglh):
    """Compute log-likelihood for all frames against one model in one orientation

    Keyword arguments:
    data_vals -- values of data intensity
    data_idx -- indexes for data values against model_slice values
    data_scale -- scale of individual frames in data
    model_slice -- slice of one model in one orientation
    out_loglh -- result will be written here
    """
    n_frames = data_vals.shape[0]
    model_sum = model_slice.sum()
    for k in range(n_frames):
        out_loglh[k] = frame_loglh(data_vals[k], data_idx[k], data_scale[k], model_slice) - \
                        model_sum*data_scale[k]



@njit(fastmath=True)
def frame_loglh(frame_vals, frame_idx, frame_scale, model_slice):
    """Compute log-likelihood for one frames against one model in one orientation

    Keyword arguments:
    frame_vals -- values of frame intensity
    frame_idx -- indexes for frame values against model_slice values
    frame_scale -- scale of a frame
    model_slice -- slice of one model in one orientation

    Return:
    loglh -- computed value of log-likelihood
    """
    frame_len = frame_vals.shape[0]
    loglh = 0
    for i in range(frame_len):
        if frame_vals[i] == 0:
            break
        loglh += frame_vals[i]*np.log(model_slice[frame_idx[i]]*frame_scale)

    return loglh


@njit(parallel=True, fastmath=True)
def maximise_data_chunk(data_vals, data_idx, data_scale,
                        q_vals, t_vals, prob_vals,
                        num_class, rotations,
                        next_model, next_model_norm):
    """Add data values to model based on probability values

    Keyword arguments:
    data_vals -- values of data intensity
    data_idx -- indexes for data values against model_slice values
    data_scale -- scale of individual frames in data
    q_vals -- radius values
    t_vals -- angle values
    prob_vals -- probability values
    num_class -- number of classes (number of models)
    rotations -- array with values of angles for rotations
    next_model -- sum of all intensities with prob_vals weights will be written here
    next_model_norm -- sum of all prob_vals weights will be written here
    """
    num_rot = len(rotations)
    for k in prange(num_class):
        new_slice = np.zeros(q_vals.shape)
        new_slice_norm = np.zeros(q_vals.shape)
        for j in range(num_rot):
            rot = rotations[j]
            if np.amax(prob_vals[k, j]) == 0:
                continue

            new_slice[:] = 0
            new_slice_norm[:] = 0

            compute_new_slice(data_vals, data_idx, data_scale,
                              prob_vals[k, j], new_slice, new_slice_norm)

            project_add_to_model(new_slice, q_vals, t_vals + rot, next_model[k])
            project_add_to_model(new_slice_norm, q_vals, t_vals + rot, next_model_norm[k])


@njit(fastmath=True)
def compute_new_slice(data_vals, data_idx, data_scale,
                      prob_vals, new_slice, new_slice_norm):
    """Compute one orientation to add data values to model based on probability values

    Keyword arguments:
    data_vals -- values of data intensity
    data_idx -- indexes for data values against model_slice values
    data_scale -- scale of individual frames in data
    prob_vals -- probability values
    new_slice -- sum of all intensities with prob_vals weights will be written here
    new_slice_norm -- sum of all prob_vals weights will be written here
    """
    n_frames, frame_len = data_vals.shape
    new_slice[:] = 0
    slice_prob = 0
    for k in range(n_frames):
        prob = prob_vals[k]
        if prob > 0:
            for i in range(frame_len):
                if data_vals[k, i] == 0:
                    break
                new_slice[data_idx[k, i]] += data_vals[k, i] * prob
            slice_prob += prob * data_scale[k]

    new_slice_norm[:] = slice_prob


@njit(parallel=True, fastmath=True)
def project_add_to_model(frame_vals, q_vals, t_vals, out_model):
    """Add values to model by polar coordinates

    Keyword arguments:
    frame_vals -- values of data intensity
    q_vals -- radius values
    t_vals -- angle values
    out_model -- result will be added to this array inplace
    """
    assert len(frame_vals) == len(q_vals)
    assert len(q_vals) == len(t_vals)

    center = (out_model.shape[0]-1) // 2

    for i in prange(len(frame_vals)):
        x_val = q_vals[i]*np.cos(t_vals[i]) + center
        y_val = q_vals[i]*np.sin(t_vals[i]) + center

        x_fl = int(np.floor(x_val))
        y_fl = int(np.floor(y_val))

        x_frac = x_val - x_fl
        y_frac = y_val - y_fl

        if test_valid(y_fl, out_model.shape[0]) and test_valid(x_fl, out_model.shape[1]):
            out_model[y_fl, x_fl] += (1 - x_frac)*(1 - y_frac)*frame_vals[i]

        if test_valid(y_fl, out_model.shape[0]) and test_valid(x_fl + 1, out_model.shape[1]):
            out_model[y_fl, x_fl + 1] += (x_frac)*(1 - y_frac)*frame_vals[i]

        if test_valid(y_fl + 1, out_model.shape[0]) and test_valid(x_fl, out_model.shape[1]):
            out_model[y_fl + 1, x_fl] += (1 - x_frac)*(y_frac)*frame_vals[i]

        if test_valid(y_fl + 1, out_model.shape[0]) and test_valid(x_fl + 1, out_model.shape[1]):
            out_model[y_fl + 1, x_fl + 1] += (x_frac)*(y_frac)*frame_vals[i]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Create PDF report for EM class data
Author: Sergey Bobkov
"""

import os
import argparse

from spi_em_class import em_data, pdfreport


def main():
    parser = argparse.ArgumentParser(description='Create PDF report for EM class data')
    parser.add_argument('-o', '--out', default=pdfreport.PDF_REPORT_NAME, help="Output file")
    parser.add_argument('-d', '--data', default=em_data.DEFAULT_DATA,
                        help='File with EM class data')
    parser_args = parser.parse_args()

    data_file = parser_args.data
    report_file = parser_args.out

    if not os.path.isfile(data_file):
        parser.error("File {} doesn't exist".format(data_file))

    pdfreport.create_pdf_report(data_file, report_file)


if __name__ == '__main__':
    main()



"""
Functions for classification log
Author: Sergey Bobkov
"""

import sys

LOG_FILENAME = 'EM.log'


def show_message(text):
    sys.stderr.write(text+'\n')
    sys.stderr.flush()


def write_log_header(param, num_frames):
    with open(LOG_FILENAME, 'w') as logfile:
        logfile.write('Classification with the EM algorithm\n\n')

        logfile.write('Data parameters:\n')
        logfile.write('\tnum data = {}\n'.format(num_frames))
        logfile.write('\tq_min = {}\n'.format(param['q_min']))
        logfile.write('\tq_max = {}\n'.format(param['q_max']))
        logfile.write('\tnum rotations = {}\n'.format(param['num_rot']))
        logfile.write('\tnum classes = {}\n'.format(param['num_class']))
        logfile.write('\tfriedel = {}\n'.format(param['friedel']))
        logfile.write('\tlogscale = {}\n'.format(param['logscale']))
        logfile.write('\tbest_proba = {}\n'.format(param['best_proba']))
        logfile.write('\tbinning = {}\n'.format(param['binning']))
        logfile.write('\n')

        logfile.write('Iter \ttime \tmutual_info\n')


def write_log_iter(iter_num, iter_time, mutual_info):
    with open(LOG_FILENAME, 'a+') as logfile:
        logfile.write('{:>4d}\t{:0.2f}\t{:0.2f}\n'.format(iter_num, iter_time, mutual_info))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Reset EM classification data
Author: Sergey Bobkov
"""

import os
import argparse

from spi_em_class import em_data


def main():
    parser = argparse.ArgumentParser(description='Reset EM classification data')
    parser.add_argument('-d', '--data', default=em_data.DEFAULT_DATA,
                        help='File with EM class data')
    parser_args = parser.parse_args()

    data_file = parser_args.data

    if not os.path.isfile(data_file):
        parser.error("File {} doesn't exist".format(data_file))

    em_data.reset_data(data_file)


if __name__ == '__main__':
    main()



"""
Code for PDF reports
Author: Sergey Bobkov
"""

import datetime
from typing import Optional
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.colors as colors
from mpl_toolkits.axes_grid1 import make_axes_locatable

from . import em_data, log_writer

mpl.rcParams['image.cmap'] = 'viridis'
mpl.rcParams['image.interpolation'] = 'nearest'

PDF_REPORT_NAME = 'em_class_report.pdf'

class ReportWriter():
    """Produce PDF reports with matplotlib figures"""
    def __init__(self, filename: str):
        self._create_file(filename)

    def _create_file(self, filename: str):
        self.pdf_pages = PdfPages(filename)
        pdf_dict = self.pdf_pages.infodict()
        pdf_dict['CreationDate'] = datetime.datetime.today()
        pdf_dict['ModDate'] = datetime.datetime.today()


    def save_figure(self, figure: plt.Figure):
        """Add a page to PDF with matplotlib figure

        Keyword arguments:
        figure -- figure to be saved
        """

        self.pdf_pages.savefig(figure)
        plt.close(figure)


    def close(self):
        """Closes PDF document"""
        self.pdf_pages.close()


def plot_image(image: np.ndarray, axis: plt.Axes, logscale: bool = True,
               colorbar: bool = True, vmin: Optional[float] = None, vmax: Optional[float] = None,
               **kwargs):
    """Plot image to axis with optional logscale and colorbar

    Keyword arguments:
    image -- 2D array with data to show
    axis -- matplotlib axis to plot
    logscale -- use logscale
    colorbar -- add colorbar to image
    vmin -- minimal value of colormap
    vmax -- miximum value of colormap
    kwargs -- arguments for imshow
    """
    if vmin is None and logscale:
        vmin = max(1, np.amin(image))
    elif vmin is None:
        vmin = np.amin(image)

    if vmax is None:
        vmax = np.amax(image)

    if logscale:
        norm = colors.LogNorm(vmin=vmin, vmax=vmax)
    else:
        norm = colors.Normalize(vmin=vmin, vmax=vmax)

    im_handle = axis.imshow(image, norm=norm, **kwargs)
    axis.set_xticks([])
    axis.set_yticks([])

    if colorbar:
        divider = make_axes_locatable(axis)
        cb_axis = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im_handle, cax=cb_axis)


def em_summary_fig(rotations, class_numbers, mutual_info):
    """Create figure with results summary

    Keyword arguments:
    rotations -- 2D array with rotation distribution over iterations
    class_numbers -- 2D array with class distribution over iterations
    mutual_info -- array with mutual info history
    """
    fig, axes = plt.subplots(1, 3, figsize=(12, 6))
    rot_axis, class_axis, info_axis = axes.flatten()

    sorted_rots = rotations[:, rotations[-1].argsort()]
    rot_axis.imshow(sorted_rots.T)
    rot_axis.set_yticks([])
    rot_axis.set_aspect('auto', 'box')
    rot_axis.set_xlabel('Iteration')
    rot_axis.set_ylabel('Pattern number')
    rot_axis.set_title('Most likely orientations of data')

    sorted_class = class_numbers[:, class_numbers[-1].argsort()]
    class_axis.imshow(sorted_class.T)
    class_axis.set_yticks([])
    class_axis.set_aspect('auto', 'box')
    class_axis.set_xlabel('Iteration')
    class_axis.set_ylabel('Pattern number')
    class_axis.set_title('Most likely classes of data')

    info_axis.plot(np.arange(1, mutual_info.size), mutual_info[1:])
    info_axis.set_aspect('auto', 'box')
    info_axis.set_xlabel('Iteration')
    info_axis.set_ylabel('I(K,W)')

    fig.tight_layout()
    return fig


def em_progress_fig(iter_num, model_data, class_numbers, rotations):
    """Create figure with iteration results

    Keyword arguments:
    iter_num -- index of iteration for title
    model_data -- 3D array with model images
    class_numbers -- array with class distribution
    rotations -- array with rotation distribution
    """
    n_classes = model_data.shape[0]
    n_rot = rotations.max()
    class_ids = np.arange(n_classes) + 1
    class_counts = [np.sum(class_numbers == i) for i in class_ids]

    plot_x = int(np.ceil(np.sqrt(n_classes)))
    plot_y = int(np.ceil(n_classes / plot_x))
    rot_bins = np.arange(n_rot)

    fig, axes = plt.subplots(plot_y, plot_x, figsize=(plot_x*3, plot_y*3))
    plt.subplots_adjust(left=0.05, bottom=0.05, right=0.90, top=0.85, wspace=0.2, hspace=0.2)
    fig.suptitle('Iteration {}'.format(iter_num), fontsize=20)

    for i in range(plot_x*plot_y):
        axis = axes.flatten()[i]
        if i >= n_classes:
            axis.set_axis_off()
        else:
            class_model = model_data[i]
            class_rotations = rotations[class_numbers == i + 1]
            vmin = class_model.mean() * 0.01
            if vmin > 0:
                plot_image(class_model, axis, vmin=vmin)
            else:
                plot_image(class_model, axis, logscale=False)

            if (class_rotations.size == 0) or (np.std(class_rotations) == 0):
                continue

            ax2 = axis.twinx()
            hist, _ = np.histogram(class_rotations, bins=rot_bins)
            # rot_unev = np.round(hist.max()/hist.mean(), 2)
            x_values = rot_bins[:-1] * (0.95 * class_model.shape[0] / rot_bins[-1])
            ax2.bar(x_values, hist, color='r', width=1, alpha=0.5)
            ax2.set_axis_off()
            ax2.set_ylim([0, hist.max()*2])
            axis.set_frame_on(False)

            axis.set_title('№{}, {} frames'.format(class_ids[i], class_counts[i]))

    return fig


def create_pdf_report(data_file, pdf_file):
    """Create figure with iteration results

    Keyword arguments:
    data_file -- file with EM class data
    pdf_file -- output pdf file
    """
    rotations = []
    classes = []
    mutual_info = []

    num_completed = em_data.get_num_saved_iterations(data_file)
    if num_completed == 0:
        log_writer.show_message('No iterations data in file: {}'.format(data_file))
        return

    for i in range(1, num_completed + 1):
        it_data = em_data.load_iteration_data(data_file, i, for_report=True)
        rotations.append(it_data['rotations'][:])
        classes.append(it_data['classes'][:])
        mutual_info.append(it_data['mutual_info'][()])

    rotations = np.stack(rotations)
    classes = np.stack(classes)
    mutual_info = np.stack(mutual_info)

    report = ReportWriter(pdf_file)
    report.save_figure(em_summary_fig(rotations, classes, mutual_info))

    it_data = em_data.load_iteration_data(data_file, num_completed)
    report.save_figure(em_progress_fig(num_completed, it_data['model'],
                                       it_data['classes'], it_data['rotations']))
    report.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Select class by rotational symmetry and save selection into CXI file
Author: Sergey Bobkov
"""

import os
import argparse
import shutil
import subprocess
import numpy as np

from spi_em_class import config, em_data, symmetry


def main():
    parser = argparse.ArgumentParser(description='Select class by rotational symmetry and save selection into CXI file')
    parser.add_argument('order', type=int, help='Symmetry order')
    parser.add_argument('-f', '--fringe', default=0, type=int, help='Fringe to analyse, 0 - consider all data')
    parser.add_argument('-t', '--threshold', default=0.5, type=float, help='Selection threshold')
    parser.add_argument('-o', '--out',
                        help="Output CXI file (default: add selection into base CXI file)")
    parser.add_argument('-c', '--class', dest='class_dset', default='em_class/classes',
                        help="Dataset with result classes")
    parser.add_argument('-s', '--select', dest='select_dset', default='em_class/select',
                        help="Dataset with result selection")
    parser.add_argument('-d', '--data', default=em_data.DEFAULT_DATA,
                        help='File with EM class data')
    parser_args = parser.parse_args()

    fringe = parser_args.fringe
    symmetry_order = parser_args.order
    symmetry_thresh = parser_args.threshold
    out_file = parser_args.out
    class_dset = parser_args.class_dset
    select_dset = parser_args.select_dset
    data_file = parser_args.data

    if not os.path.isfile(data_file):
        parser.error("File {} doesn't exist".format(data_file))

    if not 0 < symmetry_order < 50:
        parser.error("Symmetry order {} do not fit into [2, 50]".format(symmetry_order))

    if not 0 < symmetry_thresh < 1:
        parser.error("Threshold {} do not fit into (0, 1)".format(symmetry_thresh))

    params = config.load_config(data_file)
    last_iter = em_data.get_num_saved_iterations(data_file)
    if last_iter == 0:
        parser.error('Empty classification data')

    if out_file is None:
        out_file = params['cxi_file']
    else:
        if os.path.exists(out_file):
            parser.error("File {} already exist".format(out_file))

        out_dir = os.path.dirname(out_file)
        if out_dir:
            os.makedirs(out_dir, exist_ok=True)

        shutil.copy(params['cxi_file'], out_file)
        subprocess.call(["/bin/chmod", "u+w", out_file])

    result_data = em_data.load_iteration_data(data_file, last_iter)
    class_data = result_data['classes']
    model_data = result_data['model']

    if params['logscale']:
        model_data = np.exp(model_data) - 1

    good_class = []
    for i, data in enumerate(model_data):
        scores = symmetry.get_symmetry_scores(data)
        if fringe > 0:
            if fringe > scores.shape[0]:
                raise ValueError('No data for selected fringe: ' + \
                                 '{} > {}'.format(fringe, scores.shape[0] - 1))
            fringe_scores = scores[fringe]
            fringe_scores /= scores[0, :]
        else:
            fringe_scores = scores[:].sum(axis=0)

        fringe_scores[0] = 0
        fringe_scores /= fringe_scores.sum()
        order_score = fringe_scores[symmetry_order]
        if order_score >= symmetry_thresh:
            good_class.append(i + 1)
            sign = ">="
        else:
            sign = "<"

        print('Class {:2d}: {:0.2f} {} {:0.2f}'.format(i + 1, order_score, sign, symmetry_thresh))

    select_data = np.isin(class_data, [x for x in good_class])

    print('Saved selection of {}/{} frames in {} classes'.format(select_data.sum(),
                                                                 len(select_data),
                                                                 len(good_class)))

    em_data.save2cxi(out_file, class_data, class_dset)
    em_data.save2cxi(out_file, select_data, select_dset)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Run EM classification
Author: Sergey Bobkov
"""

import os
import argparse

from spi_em_class import Classifier, em_data


def main():
    parser = argparse.ArgumentParser(description='Classify CXI data by EMC')
    parser.add_argument('iter', type=int, help="Number of iterations")
    parser.add_argument('-d', '--data', default=em_data.DEFAULT_DATA,
                        help='File with EM class data')
    parser_args = parser.parse_args()

    n_iter = parser_args.iter
    data_file = parser_args.data

    if not os.path.isfile(data_file):
        parser.error("File {} doesn't exist".format(data_file))

    if n_iter < 1:
        parser.error("Number of iterations should be >= 1")

    em_class = Classifier(data_file)
    em_class.run(n_iter)


if __name__ == '__main__':
    main()



"""
Functions for configuration file
Author: Sergey Bobkov
"""

import os
import sys
import errno
import configparser
import h5py


CONFIG_PARAMS = {
    # Param names in config with corresponding types
    'best_proba': bool,
    'binning': int,
    'cxi_file': str,
    'friedel': bool,
    'logscale': bool,
    'num_class': int,
    'num_rot': int,
    'q_max': int,
    'q_min': int,
}


def save_config(data_file, param_dict, verbose=False):
    """Save parameters into hdf5 data file, update if already present

    Keyword arguments:
    data_file -- hdf5 file to save configuration
    param_dict -- dict with parameters
    verbose -- print updates to stderr
    """

    with h5py.File(data_file, 'a') as h5file:
        config_group = h5file.require_group('config')

        if verbose:
            print('Config update:', file=sys.stderr)
        for key in CONFIG_PARAMS:
            value = param_dict[key]
            if value is None:
                raise ValueError('Value of {} is None'.format(key))
            if key in config_group:
                if config_group[key][()] != value and verbose:
                    print('\tUpdate {}: {}->{}'.format(key, config_group[key][()], value),
                          file=sys.stderr)
                del config_group[key]
            elif verbose:
                print('\tAdd {}: {}'.format(key, value), file=sys.stderr)
            config_group[key] = value


def load_config(data_file):
    """Load parameters from hdf5 data file and returns them as dict

    Keyword arguments:
    data_file -- hdf5 file with configuration

    Return:
    param_dict -- dict with parameters
    """

    if not os.path.exists(data_file):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), data_file)

    with h5py.File(data_file, 'r') as h5file:
        config_group = h5file['config']
        param_dict = {}
        for key in CONFIG_PARAMS:
            param_dict[key] = config_group[key][()]

        return param_dict


def load_text_config(text_file):
    """Load parameters from text configuration file and returns them as dict

    Keyword arguments:
    text_file -- text file with configuration

    Return:
    param -- dict with parameters
    """
    config = configparser.ConfigParser()
    config.read(text_file)

    def_section = config['default']

    param = {}

    for key, key_type in CONFIG_PARAMS.items():
        if key_type == bool:
            param[key] = def_section.getboolean(key)
        elif key_type == int:
            param[key] = def_section.getint(key)
        elif key_type == str:
            param[key] = def_section.get(key)
        else:
            raise ValueError('Unknown parameter type: {}'.format(key_type))

    return param


def save_text_config(text_file, param):
    """Save parameters into text file

    Keyword arguments:
    text_file -- text file to save configuration
    param -- dict with parameters
    """
    config = configparser.ConfigParser()
    def_sect = {}
    for key, key_type in CONFIG_PARAMS.items():
        if key_type == bool:
            def_sect[key] = 'yes' if param[key] else 'no'
        else:
            def_sect[key] = param[key]

    config['default'] = def_sect

    with open(text_file, 'w') as configfile:
        config.write(configfile)


def default_config():
    """Load default config

    Return:
    param -- dict with parameters
    """
    config_file = os.path.join(os.path.dirname(__file__), 'data/default_config.ini')
    return load_text_config(config_file)



"""
EM Classifier
Author: Sergey Bobkov
"""

import numpy as np

from . import config, em_compute, em_data, log_writer, timer

class Classifier:
    """Class that perform EM classification"""
    def __init__(self, data_file, chunk_size=1000):
        params = config.load_config(data_file)

        self.data_file = data_file
        self.cxi_file = params['cxi_file']
        self.q_min = params['q_min']
        self.q_max = params['q_max']
        self.num_rot = params['num_rot']
        self.num_class = params['num_class']
        self.friedel = params['friedel']
        self.logscale = params['logscale']
        self.best_proba = params['best_proba']
        self.binning = params['binning']
        self.chunk_size = chunk_size

        self.mutual_info = []
        self.q_vals = [] # list[1d np.ndarray]
        self.t_vals = [] # list[1d np.ndarray]
        self.group_names = None
        self.num_frames = []
        self.data_scale = None
        self.rotations = None
        self.model = None
        self.loglh = None
        self.prob_vals = None
        self.next_model = None

        self._calculate_rotations()
        self._load_converted_data()
        self._init_class_arrays()
        self._load_finished_iterations()

    def _calculate_rotations(self):
        self.rotations = np.linspace(0,
                                     np.pi if self.friedel else 2*np.pi,
                                     self.num_rot,
                                     endpoint=False)

    def _load_converted_data(self):
        convert_params = em_data.load_convert_params(self.data_file)
        if convert_params is None:
            log_writer.show_message('No cached data, loading from CXI')
            em_data.convert_cxi_data(self.data_file)
            convert_params = em_data.load_convert_params(self.data_file)

        if convert_params['q_max'] != self.q_max or \
                convert_params['q_min'] != self.q_min or \
                convert_params['logscale'] != self.logscale or \
                convert_params['binning'] != self.binning:
            raise ValueError('Cashed data do not fit with current parameters.\n' + \
                             'Please reset classification.')

        self.group_names = em_data.get_converted_group_names(self.data_file)

        self.num_frames = []
        for name in self.group_names:
            self.q_vals.append(em_data.load_q_values(self.data_file, name))
            self.t_vals.append(em_data.load_t_values(self.data_file, name))
            self.num_frames.append(em_data.get_num_frames(self.data_file, name))


    def _init_class_arrays(self):
        self.data_scale = np.zeros(sum(self.num_frames))
        self.model = 0.99 + 0.01*np.random.rand(*self._optimal_model_shape())
        self.loglh = np.zeros((self.num_class, self.num_rot, sum(self.num_frames)))
        self.prob_vals = self.loglh.copy()
        self.prob_vals[0, 0, :] = 1 # For fast scaling at 1 iteration
        self.next_model = np.zeros(self.model.shape)


    def _load_finished_iterations(self):
        num_completed = em_data.get_num_saved_iterations(self.data_file)

        if num_completed > 0:
            it_data = em_data.load_iteration_data(self.data_file, num_completed)
            it_model_shape = it_data['model'].shape

            if self._optimal_model_shape() == it_model_shape:
                self._set_model(it_data['model'])
                self._set_loglh(it_data['loglh'])
            else:
                raise ValueError('EM parameters do not match the last iteration.\n' + \
                                 'Please reset classification.')


    def _save_iteration(self):
        it_data = {
            'model': self.model,
            'rotations': np.argmax(self.prob_vals.max(axis=0), axis=0),
            'classes': np.argmax(self.prob_vals.max(axis=1), axis=0) + 1,
            'scale': self.data_scale,
            'mutual_info': self._get_mutual_info(),
            'loglh': self.loglh
        }
        em_data.save_iteration_data(self.data_file, it_data)


    def _optimal_model_shape(self):
        q_max = max(q_vals.max() for q_vals in self.q_vals)
        model_side = 2*int(np.ceil(q_max)) + 1

        return (self.num_class, model_side, model_side)


    def _set_model(self, model):
        assert model.shape == self._optimal_model_shape()
        assert model.min() >= 0

        self.model[:] = model


    def _set_loglh(self, loglh):
        assert loglh.shape == self.loglh.shape, (loglh.shape, self.loglh.shape)
        assert not np.isnan(loglh).any()

        self.loglh[:] = loglh
        self._calculate_prob()


    def run(self, num_iter):
        """Perform num_iter EM iterations

        Keyword arguments:
        num_iter -- Number of iterations to perform
        """
        params = config.load_config(self.data_file)
        num_completed = em_data.get_num_saved_iterations(self.data_file)
        if num_completed == 0:
            log_writer.write_log_header(params, sum(self.num_frames))

        for i in range(num_completed + 1, num_completed + num_iter + 1):
            log_writer.show_message('Iteration {}'.format(i))
            with timer.Timer() as iter_tmr:
                self.run_iteration()

            log_writer.write_log_iter(i, iter_tmr.interval, self._get_mutual_info())

        log_writer.show_message('{} iterations completed'.format(num_iter))


    def run_iteration(self):
        """Perform one EM iteration"""
        self._prepare_iteration()

        with timer.Timer() as tmr:
            self._expect()
        log_writer.show_message('\tCalculating probabilities in {:0.2f} sec.'.format(tmr.interval))

        with timer.Timer() as tmr:
            self._maximise()
        log_writer.show_message('\tUpdate model in {:0.2f} sec.'.format(tmr.interval))

        self._set_model(self.next_model.copy())
        self._save_iteration()


    def _prepare_iteration(self):
        # Replace zeros in model with lowest possible number. We need that to calculate log()
        self.model[self.model == 0] = np.finfo(self.model.dtype).tiny


    def _expect(self):
        for i, group_frames in enumerate(self.num_frames):
            group_name = self.group_names[i]
            part_start = sum(self.num_frames[:i])
            part_end = part_start + group_frames
            q_vals = self.q_vals[i]
            t_vals = self.t_vals[i]
            data_scale = self.data_scale[part_start:part_end]

            model_slice_buf = np.zeros((self.num_class, self.num_rot, len(q_vals)), dtype='d')
            loglh_buf = np.zeros((self.num_class, self.num_rot, group_frames))

            em_compute.extract_all_model(self.model, q_vals, t_vals, self.rotations,
                                         model_slice_buf)

            group_prob = self.prob_vals[:, :, part_start:part_end].reshape((-1, group_frames))
            model_slice_sum = np.sum(model_slice_buf, axis=2).flatten()
            frame_model_norm = np.dot(group_prob.T, model_slice_sum)

            for start in range(0, group_frames, self.chunk_size):
                end = min(start + self.chunk_size, group_frames)
                data_vals, data_idx = em_data.load_data_values(self.data_file, group_name,
                                                               start, end)

                data_scale[start:end] = data_vals.sum(axis=1)/frame_model_norm[start:end]

                em_compute.compute_all_loglh(data_vals, data_idx,
                                             data_scale[start:end],
                                             model_slice_buf,
                                             loglh_buf[:, :, start:end])

            self.loglh[:, :, part_start:part_end] = loglh_buf

            del model_slice_buf
            del frame_model_norm
            del loglh_buf

        self.loglh = np.nan_to_num(self.loglh) # Fix -inf
        self._calculate_prob()


    def _calculate_prob(self):
        loglh = self.loglh - self.loglh.max(axis=(0, 1))
        if self.best_proba:
            self.prob_vals = (loglh == 0).astype(float)

            # Fix classes with zero frames
            classes = np.argmax(self.prob_vals.max(axis=1), axis=0)
            class_counts = np.zeros(self.num_class)
            for i in range(self.num_class):
                class_counts[i] = np.sum(classes == i)

            if class_counts.min() == 0:
                for i in np.nonzero(class_counts == 0)[0]:
                    self.prob_vals[i] = np.exp(loglh[i])

                self._normalise_prob()
        else:
            self.prob_vals = np.exp(loglh)
            self._normalise_prob()


    def _normalise_prob(self):
        self.prob_vals /= self.prob_vals.sum(axis=(0, 1))


    def _maximise(self):
        # Normalise scale
        self.data_scale /= self.data_scale.mean()

        next_model = np.zeros_like(self.next_model)
        next_model_norm = np.zeros_like(self.next_model)

        for i, group_frames in enumerate(self.num_frames):
            group_name = self.group_names[i]
            part_start = sum(self.num_frames[:i])
            part_end = part_start + group_frames
            q_vals = self.q_vals[i]
            t_vals = self.t_vals[i]
            data_scale = self.data_scale[part_start:part_end]
            prob_vals = self.prob_vals[:, :, part_start:part_end]
            # blacklist_chunk = self.blacklist[part_start:part_end]

            for start in range(0, group_frames, self.chunk_size):
                end = min(start + self.chunk_size, group_frames)
                data_vals, data_idx = em_data.load_data_values(self.data_file, group_name,
                                                               start, end)

                em_compute.maximise_data_chunk(data_vals, data_idx,data_scale[start:end],
                                               q_vals, t_vals, prob_vals[:, :, start:end],
                                               self.num_class, self.rotations,
                                               next_model, next_model_norm)

        self.next_model.fill(0)
        mask = next_model_norm > 0
        self.next_model[mask] = next_model[mask] / next_model_norm[mask]


    def _get_mutual_info(self):
        return np.sum(self.prob_vals*(self.loglh + np.log(self.num_rot)))/sum(self.num_frames)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Configure EM classification parameters
Author: Sergey Bobkov
"""

import os
import sys
import argparse
import configparser
import tempfile
import subprocess

from spi_em_class import config, em_data


def get_error_lines(param):
    err_lines = []

    if 'cxi_file' in param:
        if param['cxi_file'] is None:
            err_lines.append("cxi_file is not defined")
        elif not os.path.exists(param['cxi_file']) or not os.path.isfile(param['cxi_file']):
            err_lines.append("File {} doesn't exist".format(param['cxi_file']))
        elif not em_data.check_cxi_file(param['cxi_file']):
            err_lines.append("{} is not a valid cxi file".format(param['cxi_file']))

    if 'q_max' in param:
        if param['q_max'] is None:
            err_lines.append("q_max is not defined")
        elif param['q_max'] < 0:
            err_lines.append("Negative q_max is not allowed")

    if 'q_min' in param:
        if param['q_min'] is None:
            err_lines.append("q_min is not defined")
        elif param['q_min'] < 0:
            err_lines.append("Negative q_min is not allowed")

    if 'q_max' in param and 'q_min' in param:
        if param['q_max'] < param['q_min'] and param['q_max'] > 0:
            err_lines.append("q_max cannot be < q_min")

    if 'num_rot' in param:
        if param['num_rot'] is None:
            err_lines.append("num_rot is not defined")
        elif param['num_rot'] < 1:
            err_lines.append("Number of rotations < 1 is not allowed")

    if 'num_class' in param:
        if param['num_class'] is None:
            err_lines.append("num_class is not defined")
        elif param['num_class'] < 1:
            err_lines.append("Number of classes < 1 is not allowed")

    if 'binning' in param:
        if param['binning'] is None:
            err_lines.append("binning is not defined")
        elif param['binning'] < 1:
            err_lines.append("binning < 1 is not allowed")

    return err_lines


def edit_params(data_file):
    param = config.load_config(data_file)
    tmpf = tempfile.NamedTemporaryFile(suffix=".tmp", delete=False)
    tmpf.close()
    config.save_text_config(tmpf.name, param)

    while True:
        subprocess.call([os.environ.get('EDITOR', 'mcedit'), tmpf.name])
        try:
            new_param = config.load_text_config(tmpf.name)
        except (configparser.Error, ValueError):
            with open(tmpf.name, "a+") as openfile:
                openfile.write('# Parsing Error, please fix and save file\n')
                continue

        err_lines = get_error_lines(new_param)
        if err_lines:
            config.save_text_config(tmpf.name, new_param)
            with open(tmpf.name, "a+") as openfile:
                openfile.write('# ERROR:')
                for line in err_lines:
                    openfile.write('# ' + line + '\n')
                openfile.write('# Invalid parameters, please fix and save file\n')
                continue

        break

    os.remove(tmpf.name)
    config.save_config(data_file, new_param, verbose=True)


def main():
    parser = argparse.ArgumentParser(description='Configure EM classification parameters')
    parser.add_argument('-d', '--data', default=em_data.DEFAULT_DATA,
                        help='File with EM class data')
    parser.add_argument('--cxi', dest='cxi_file', metavar='FILE', help='Input CXI file')
    parser.add_argument('--q_min', dest='q_min', type=int,
                        help='Minumum q-radius used in classification (pixels)')
    parser.add_argument('--q_max', dest='q_max', type=int,
                        help='Maximim q-radius used in classification (pixels)')
    parser.add_argument('--num_rot', dest='num_rot', type=int,
                        help="Number of considered rotation angles")
    parser.add_argument('--num_class', dest='num_class', type=int,
                        help="Number of EM classes")
    parser.add_argument('--friedel', dest='friedel', action='store_true',
                        help="Apply Friedel symmetry")
    parser.add_argument('--no-friedel', dest='nofriedel', action='store_true',
                        help="Do not apply Friedel symmetry")
    parser.add_argument('--logscale', dest='logscale', action='store_true',
                        help="Apply log-scaling to input data")
    parser.add_argument('--no-logscale', dest='nologscale', action='store_true',
                        help="Do not apply log-scaling to input data")
    parser.add_argument('--best', dest='best_proba', action='store_true',
                        help="Consider only one orientation with best probability for each frame")
    parser.add_argument('--no-best', dest='no_best_proba', action='store_true',
                        help="Consider all orientations")
    parser.add_argument('--binning', dest='binning', type=int,
                        help="Bin input frames (Combine pixels in BIN*BIN groups togeather)")
    parser.add_argument('-s', '--show', dest='show', action='store_true',
                        help="Show current configuration (ignore other parameters)")
    parser.add_argument('-e', '--editor', dest='editor', action='store_true',
                        help="Open editor (default if no parameters)")
    parser_args = parser.parse_args()

    data_file = parser_args.data
    show = parser_args.show
    editor = parser_args.editor

    if len(sys.argv) <= 1:
        editor = True

    if not os.path.isfile(data_file):
        parser.error("File {} doesn't exist".format(data_file))

    if editor:
        edit_params(data_file)
    elif show:
        print('Current parameters:')
        param = config.load_config(data_file)
        for key in param:
            print('{:>20s}: {}'.format(key, param[key]))
    else:
        param = config.load_config(data_file)
        if parser_args.friedel and parser_args.nofriedel:
            parser.error("Parameters --friedel and --no-friedel are not compatible")

        if parser_args.best_proba and parser_args.no_best_proba:
            parser.error("Parameters --best and --no-best are not compatible")

        if parser_args.cxi_file is not None:
            cxi_file = parser_args.cxi_file
            param['cxi_file'] = os.path.abspath(cxi_file)

        if parser_args.q_max is not None:
            param['q_max'] = parser_args.q_max

        if parser_args.q_min is not None:
            param['q_min'] = parser_args.q_min

        if parser_args.num_rot is not None:
            param['num_rot'] = parser_args.num_rot

        if parser_args.num_class is not None:
            param['num_class'] = parser_args.num_class

        if parser_args.friedel:
            param['friedel'] = True

        if parser_args.nofriedel:
            param['friedel'] = False

        if parser_args.logscale:
            param['logscale'] = True

        if parser_args.nologscale:
            param['logscale'] = False

        if parser_args.best_proba:
            param['best_proba'] = True

        if parser_args.no_best_proba:
            param['best_proba'] = False

        if parser_args.binning is not None:
            param['binning'] = parser_args.binning

        err_lines = get_error_lines(param)
        if err_lines:
            parser.error('\n'.join(err_lines))
        else:
            config.save_config(data_file, param, verbose=True)


if __name__ == '__main__':
    main()



"""
Simple class for computation timing
Author: Sergey Bobkov
"""

import time

class Timer:
    """
    Timer context manager

    Example:
        with Timer() as t:
            do_smth()

        print('Time elapsed: %.03f sec.' % t.interval)

    """
    def __init__(self):
        self.start = 0
        self.end = 0
        self.interval = 0

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.interval = self.end - self.start



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='spi_em_class',
      version='0.1.5',
      description='Module for EM classification of diffraction images in CXI format.',
      long_description=readme(),
      long_description_content_type="text/markdown",
      url='https://gitlab.com/spi_xfel/spi_em_class',
      author='Sergey Bobkov',
      author_email='s.bobkov@grid.kiae.ru',
      license='MIT',
      python_requires='>=3.6',
      install_requires=['numba',
                        'numpy',
                        'scipy',
                        'matplotlib',
                        'h5py',
                        'tqdm'],
      packages=['spi_em_class'],
      scripts=['scripts/spi_em_create.py',
               'scripts/spi_em_config.py',
               'scripts/spi_em_report.py',
               'scripts/spi_em_reset.py',
               'scripts/spi_em_run.py',
               'scripts/spi_em_save.py',
               'scripts/spi_em_save_by_symmetry.py'],
      include_package_data=True,
      zip_safe=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Create new EM classification
Author: Sergey Bobkov
"""

import os
import argparse

from spi_em_class import config, em_data


def main():
    parser = argparse.ArgumentParser(description='Create new EM classification')
    parser.add_argument('cxi_file', metavar='FILE', help='Input CXI file')
    parser.add_argument('-d', '--data', default=em_data.DEFAULT_DATA, help='Output file with EM class data')
    parser_args = parser.parse_args()

    cxi_file = parser_args.cxi_file
    data_file = parser_args.data

    if not os.path.isfile(cxi_file):
        parser.error("File {} doesn't exist".format(cxi_file))

    if not em_data.check_cxi_file(cxi_file):
        parser.error("File {} is not a valid CXI file".format(cxi_file))

    if os.path.exists(data_file):
        parser.error("File {} already exist".format(data_file))
    else:
        dirname = os.path.dirname(data_file)
        if dirname:
            os.makedirs(dirname, exist_ok=True)

    params = config.default_config()
    params['cxi_file'] = os.path.abspath(cxi_file)

    config.save_config(data_file, params)

    print('Create new EM classification data in {}'.format(data_file))


if __name__ == '__main__':
    main()



"""
Processing of image data
Author: Sergey Bobkov
"""

import os
import h5py
import numpy as np
from tqdm import tqdm

from . import config

DEFAULT_DATA = 'em_class_data.h5'
CONVERT_ROOT = 'converted_data'
CONVERT_PARAMS = ['q_max', 'q_min', 'binning', 'logscale']
ITERATION_PARAMS = ['model', 'rotations', 'classes', 'scale', 'mutual_info', 'loglh']
REPORT_PARAMS = ['rotations', 'classes', 'mutual_info']


def check_cxi_file(cxi_file):
    """Check if CXI file is present and include required data

    Keyword arguments:
    cxi_file -- CXI file

    Retuns:
    status -- True if everything correct, False otherwise
    """
    try:
        with h5py.File(cxi_file, 'r') as h5file:
            if 'entry_1/image_1/data' not in h5file:
                raise OSError
        return True
    except OSError:
        return False


def convert_cxi_data(data_file):
    """Convert data from CXI file into format suited for classification and save in in data_file

    Keyword arguments:
    data_file -- hdf5 file with classification data
    """

    params = config.load_config(data_file)

    with h5py.File(data_file, 'a') as h5file:
        if CONVERT_ROOT in h5file:
            del h5file[CONVERT_ROOT]

        convert_group = h5file.create_group(CONVERT_ROOT)
        for key in CONVERT_PARAMS:
            convert_group[key] = params[key]

        with h5py.File(params['cxi_file'], 'r') as cxi_file:
            entry = cxi_file['entry_1']
            image_keys = [k for k in entry.keys() if k.startswith('image_')]
            for k in tqdm(image_keys):
                cxi_image = entry[k]
                convert_image = convert_group.create_group(k)
                convert_image_group(cxi_image, convert_image, params)
                create_sparse_data(convert_image)
                # Remove dense data
                del convert_image['data']


def convert_image_group(cxi_image, convert_image, params, chunk_size=1000):
    """Convert image group from CXI file and save result in convert_image group

    Keyword arguments:
    cxi_image -- image group in CXI file
    convert_image -- empty group to save converted data
    params -- dict with parameters
    chunk_size -- number of frames processed at once
    """
    data_dset = cxi_image['data']
    mask = cxi_image['mask'][:]
    center = cxi_image['image_center'][:]

    assert center.shape == (3,)
    assert mask.ndim == 2
    assert data_dset.ndim == 3
    assert data_dset.shape[1:] == mask.shape

    num_frames = data_dset.shape[0]

    q_max = params['q_max']

    if q_max != 0:
        eff_width = 2*q_max + 2
        x_start = max(np.floor(center[0] - eff_width//2).astype(int), 0)
        x_end = min(np.floor(center[0] + eff_width//2).astype(int), data_dset.shape[2])
        y_start = max(np.floor(center[1] - eff_width//2).astype(int), 0)
        y_end = min(np.floor(center[1] + eff_width//2).astype(int), data_dset.shape[1])
        corr_center = np.array([center[0] - x_start, center[1] - y_start, 0])
        mask_buf = mask[y_start:y_end, x_start:x_end]
    else:
        corr_center = center
        mask_buf = mask

    for chunk_start in tqdm(range(0, num_frames, chunk_size), position=1):
        chunk_end = chunk_start + chunk_size
        if q_max != 0:
            data_buf = data_dset[chunk_start:chunk_end, y_start:y_end, x_start:x_end]
        else:
            data_buf = data_dset[chunk_start:chunk_end]

        q_values, t_values, converted_data_buf = \
            convert_data(data_buf, mask_buf, corr_center, params)

        if chunk_start == 0:
            convert_image.create_dataset('q_values', data=q_values, compression="lzf")
            convert_image.create_dataset('t_values', data=t_values, compression="lzf")
            converted_data_dset = convert_image.create_dataset(
                'data', shape=(num_frames, len(q_values)),
                chunks=(1, len(q_values)), compression="lzf")

        if params['logscale']:
            converted_data_buf = np.log(converted_data_buf + 1)

        converted_data_dset[chunk_start:chunk_end] = converted_data_buf
        del converted_data_buf


def convert_data(data, mask, center, params):
    """Convert 3D image array into set of points with polar coordinates

    Keyword arguments:
    data -- 3D image array (index, y, x)
    mask -- 2D mask for image array
    center -- image center coordinates
    params -- dict with parameters

    Return:
    q_values -- point radius in polar coordinates
    t_values -- point angles in polar coordinates
    converted_data_buf -- 2D array with intensity values (image_index, point_index)
    """

    binning = params['binning']
    q_max = params['q_max']
    q_min = params['q_min']
    friedel = params['friedel']

    if binning > 1:
        mask = (apply_binning(mask, binning) > 0)

    q_grid, t_grid = polar_coords(mask.shape, center/binning)
    q_values = q_grid[mask == 0]
    t_values = t_grid[mask == 0]

    if friedel:
        q_values = np.concatenate([q_values, q_values])
        t_values = np.concatenate([t_values, t_values + np.pi])

    q_filter = np.ones_like(q_values, dtype=bool)
    if q_max != 0:
        q_filter *= (q_values <= q_max/binning)
    if q_min != 0:
        q_filter *= (q_values >= q_min/binning)

    q_values = q_values[q_filter]
    t_values = t_values[q_filter]

    converted_data_buf = np.zeros((len(data), len(q_values)), dtype=data.dtype)

    for j, data_frame in enumerate(data):
        frame_vals = apply_binning(data_frame, binning)[mask == 0]
        if friedel:
            frame_vals = np.concatenate([frame_vals, frame_vals])

        converted_data_buf[j] = frame_vals[q_filter]

    return q_values, t_values, converted_data_buf


def polar_coords(shape, center):
    """Compute polar coordinates for grid with defined shape and center

    Keyword arguments:
    shape -- 2D grid shape
    center -- center coordinates

    Return:
    q_grid -- array of point radius in polar coordinates
    t_grid -- array of point angles in polar coordinates
    """
    assert len(shape) == 2
    assert center.shape == (3,)

    size_y, size_x = shape

    x_range = np.arange(size_x) - center[0]
    y_range = np.arange(size_y) - center[1]

    x_grid, y_grid = np.meshgrid(x_range, y_range)

    q_grid = np.sqrt(x_grid**2 + y_grid**2)
    t_grid = np.arctan2(y_grid, x_grid)

    return q_grid, t_grid


def apply_binning(data, binning):
    """Apply binning to 2D image, combine groups of x*x pixels
    and return sum of intensities in each group

    Keyword arguments:
    data -- 2D image
    binning -- size of binned group, i.e. 2 for 2 by 2 binning.

    Return:
    binned_data -- binned image
    """
    size_y, size_x = data.shape

    bin_x = size_x//binning #+ (1 if X%bin_value > bin_value//2 else 0)
    bin_y = size_y//binning #+ (1 if Y%bin_value > bin_value//2 else 0)

    binned_data = data[:bin_y*binning, :bin_x*binning]
    binned_data = binned_data.reshape(bin_y, binning, bin_x, binning).sum(axis=(1, 3))

    return binned_data


def create_sparse_data(convert_image):
    """Replace 'data' with sparse 'values' and 'indexes' datasets

    Keyword arguments:
    convert_image -- group with 'data' to process
    """
    data_dset = convert_image['data']
    num_frames = len(data_dset)
    max_npix = np.count_nonzero(data_dset, axis=1).max()

    values_dest = convert_image.create_dataset('values', dtype=data_dset.dtype,
                                               shape=(num_frames, max_npix), chunks=(1, max_npix),
                                               compression="lzf")
    index_dest = convert_image.create_dataset('indexes', dtype='i4',
                                              shape=(num_frames, max_npix), chunks=(1, max_npix),
                                              compression="lzf")

    for i, frame in enumerate(data_dset):
        idx = np.nonzero(frame > 0)[0]
        index_dest[i, :len(idx)] = idx
        values_dest[i, :len(idx)] = frame[idx]


def load_convert_params(data_file):
    """Load params of convertion for data_file

    Keyword arguments:
    data_file -- hdf5 file with classification data

    Return:
    convert_params -- dict with convert params (None if not available)
    """
    convert_params = {}
    with h5py.File(data_file, 'r') as h5file:
        if CONVERT_ROOT not in h5file:
            return None

        convert_group = h5file[CONVERT_ROOT]
        for key in convert_group:
            if key.startswith('image_'):
                image = convert_group[key]
                if 'values' not in image or 'indexes' not in image:
                    return None

        for key in CONVERT_PARAMS:
            if key not in convert_group:
                return None
            convert_params[key] = convert_group[key][()]

        return convert_params


def get_converted_group_names(data_file):
    """Get names of groups in converted data

    Keyword arguments:
    data_file -- hdf5 file with classification data

    Return:
    converted_group_names -- list of group names
    """

    with h5py.File(data_file, 'r') as h5file:
        if CONVERT_ROOT not in h5file:
            raise ValueError('Converted data is not available')

        convert_group = h5file[CONVERT_ROOT]

        converted_group_names = [name for name in convert_group.keys()
                                 if isinstance(convert_group[name], h5py.Group)]

    return converted_group_names


def get_num_frames(data_file, group_name):
    """Get number of frames in group

    Keyword arguments:
    data_file -- hdf5 file with classification data
    group_name -- name of group in converted data

    Return:
    length -- number of frames
    """

    with h5py.File(data_file, 'r') as h5file:
        if CONVERT_ROOT not in h5file:
            raise ValueError('Converted data is not available')

        convert_group = h5file[CONVERT_ROOT]
        data_group = convert_group[group_name]

        return data_group['values'].shape[0]


def _load_converted_data(data_file, group_name, dataset, start=0, end=0):
    """Load data from converted dataset

    Keyword arguments:
    data_file -- hdf5 file with classification data
    group_name -- name of group in converted data
    dataset -- name of dataset to read
    start -- start index for dataset slice
    end -- end index index for dataset slice

    Return:
    data -- loaded data
    """

    with h5py.File(data_file, 'r') as h5file:
        if CONVERT_ROOT not in h5file:
            raise ValueError('Converted data is not available')

        convert_group = h5file[CONVERT_ROOT]
        data_group = convert_group[group_name]
        data_dset = data_group[dataset]

        if end == 0:
            end = data_dset.shape[0]

        return data_dset[start:end]


def load_q_values(data_file, group_name):
    """Load 'q_values' datasets from converted data

    Keyword arguments:
    data_file -- hdf5 file with classification data
    group_name -- name of group in converted data

    Return:
    q_values -- loaded data
    """
    return _load_converted_data(data_file, group_name, 'q_values')


def load_t_values(data_file, group_name):
    """Load 'q_values' datasets from converted data

    Keyword arguments:
    data_file -- hdf5 file with classification data
    group_name -- name of group in converted data

    Return:
    q_values -- loaded data
    """
    return _load_converted_data(data_file, group_name, 't_values')


def load_data_values(data_file, group_name, start=0, end=0):
    """Load sparse data from group in converted data

    Keyword arguments:
    data_file -- hdf5 file with classification data
    group_name -- name of group in converted data
    start -- start index for dataset slice
    end -- end index index for dataset slice

    Return:
    values -- data values
    indexes -- indexes for data values
    """
    values = _load_converted_data(data_file, group_name, 'values', start, end)
    indexes = _load_converted_data(data_file, group_name, 'indexes', start, end)
    return values, indexes


def reset_data(data_file):
    """Replace data_file with new file with same config

    Keyword arguments:
    data_file -- hdf5 file with classification data
    """
    param = config.load_config(data_file)
    os.remove(data_file)
    config.save_config(data_file, param)


def load_iteration_data(data_file, iteration, for_report=False):
    """Load iteration result from data_file

    Keyword arguments:
    data_file -- hdf5 file with classification data
    iteration -- iteration number
    for_report -- load only data for report

    Return:
    it_data -- dict with iteration data
    """
    if iteration <= 0:
        raise ValueError("Iteration cannot be <= 0, got {}".format(iteration))
    elif iteration > get_num_saved_iterations(data_file):
        raise ValueError("Data for itaration {} is not available".format(iteration))

    with h5py.File(data_file, 'r') as h5file:
        it_group = h5file['iter_{}'.format(iteration)]
        it_data = {}
        for k in ITERATION_PARAMS:
            if for_report and k not in REPORT_PARAMS:
                continue
            it_data[k] = it_group[k][...]

        return it_data


def get_num_saved_iterations(data_file):
    """Return number of saved iterations in data_file

    Keyword arguments:
    data_file -- hdf5 file with classification data

    Return:
    iter_num -- number of saved iterations
    """
    with h5py.File(data_file, 'r') as h5file:
        iter_numbers = []
        for k in h5file.keys():
            if k.startswith('iter_'):
                iter_numbers.append(int(k[5:]))

        if not iter_numbers:
            return 0

        return max(iter_numbers)


def save_iteration_data(data_file, it_data):
    """Save iteration result to data_file

    Keyword arguments:
    data_file -- hdf5 file with classification data
    it_data -- dict with iteration data

    Return:
    iteration -- number of saved iteration
    """
    iteration = get_num_saved_iterations(data_file) + 1
    with h5py.File(data_file, 'a') as h5file:
        it_group = h5file.require_group('iter_{}'.format(iteration))

        for k in ITERATION_PARAMS:
            it_group[k] = it_data[k]

    return iteration


def save2cxi(cxi_file, data, dset_name):
    """Save classification result into CXI file

    Keyword arguments:
    cxi_file -- CXI file
    data -- 1D array with data, length should be equal to number of frames in CXI file
    dset_name -- name for result dataset within image_n group
    """

    start = 0

    with h5py.File(cxi_file, 'a') as cxi_file:
        entry = cxi_file['entry_1']
        for k in entry.keys():
            if k.startswith('image_'):
                cxi_image = entry[k]
                if dset_name in cxi_image:
                    del cxi_image[dset_name]

                num_frames = cxi_image['data'].shape[0]
                end = start + num_frames
                cxi_image[dset_name] = data[start:end]
                start += num_frames

    if end != len(data):
        raise ValueError('Data lenght not equal to frames number: {} != {}'.format(end, len(data)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Save selection into CXI file
Author: Sergey Bobkov
"""

import os
import argparse
import shutil
import subprocess
import numpy as np

from spi_em_class import config, em_data


def main():
    parser = argparse.ArgumentParser(description='Save selection into CXI file')
    parser.add_argument('select_class', nargs='+', type=int, help='Selected classes')
    parser.add_argument('-o', '--out',
                        help="Output CXI file (default: add selection into base CXI file)")
    parser.add_argument('-c', '--class', dest='class_dset', default='em_class/classes',
                        help="Dataset with result classes")
    parser.add_argument('-s', '--select', dest='select_dset', default='em_class/select',
                        help="Dataset with result selection")
    parser.add_argument('-d', '--data', default=em_data.DEFAULT_DATA,
                        help='File with EM class data')
    parser_args = parser.parse_args()

    good_class = parser_args.select_class
    out_file = parser_args.out
    class_dset = parser_args.class_dset
    select_dset = parser_args.select_dset
    data_file = parser_args.data

    if not os.path.isfile(data_file):
        parser.error("File {} doesn't exist".format(data_file))

    params = config.load_config(data_file)

    for cla in good_class:
        if not 0 < cla < params['num_class']:
            parser.error("Class {} do not fit into [1, {}]".format(cla, params['num_class']))

    last_iter = em_data.get_num_saved_iterations(data_file)
    if last_iter == 0:
        parser.error('Empty classification data')

    if out_file is None:
        out_file = params['cxi_file']
    else:
        if os.path.exists(out_file):
            parser.error("File {} already exist".format(out_file))

        out_dir = os.path.dirname(out_file)
        if out_dir:
            os.makedirs(out_dir, exist_ok=True)

        shutil.copy(params['cxi_file'], out_file)
        subprocess.call(["/bin/chmod", "u+w", out_file])

    result_data = em_data.load_iteration_data(data_file, last_iter, for_report=True)
    class_data = result_data['classes']
    select_data = np.isin(class_data, [x for x in good_class])

    print('Saved selection of {}/{} frames in {} classes'.format(select_data.sum(),
                                                                 len(select_data),
                                                                 len(good_class)))

    em_data.save2cxi(out_file, class_data, class_dset)
    em_data.save2cxi(out_file, select_data, select_dset)


if __name__ == '__main__':
    main()



"""
Processing of image data
Author: Sergey Bobkov
"""

import numpy as np
import scipy as sp
from scipy.fftpack import dct
from scipy.ndimage import map_coordinates
from scipy.optimize import differential_evolution
from scipy.signal import argrelextrema


def sphere_form_factor(x_val, sphere_r, scale):
    """Calculate sphere form factor

    Keyword arguments:
    x -- spatial frequency
    r -- sphere radius
    scale -- intensity multiplier

    Return:
    res -- sphere form factor value
    """

    q_by_r = x_val*2*np.pi*sphere_r
    vol = (4/3)*np.pi*(sphere_r**3)
    with np.errstate(divide='ignore', invalid='ignore'):
        res = (scale/vol)*((3*vol*(np.sin(q_by_r) - q_by_r*np.cos(q_by_r))/q_by_r**3)**2)
    return np.nan_to_num(res)


def sphere_size(first_min: int, freq_step: float):
    """Compute size of a sphere by position of first min in sphere form factor

    Keyword arguments:
    first_min_pix -- position of first minimum of form factor (pix)
    freq_step -- Spatial frequency for angle of one detector pixel (1/Angstrom)

    Return:
    size -- sphere size (pixel_size/(wavelength*distance)) (Angstrom)
    """

    return 4.5 / (np.pi*first_min*freq_step)


def get_psd_minima(psd_data):
    """Retrieve index of minima in PSD data by fitting of spherical form-factor

    Keyword arguments:
    psd_data -- array of PSD values, starting from 0 angle. 0 values treated as masked data.

    Return:
    min_values -- indexes of minima in psd_data
    """
    size_min = sphere_size(psd_data.shape[0], 1)
    size_max = sphere_size(6, 1)
    bounds = [(size_min, size_max)]
    x_coord = np.arange(len(psd_data))[psd_data > 0]
    data_to_fit = psd_data[psd_data > 0]

    def r_qual(x):
        fit = sphere_form_factor(x_coord, x, 1)
        fit *= data_to_fit.mean()/fit.mean()
        return np.abs(data_to_fit - fit).sum()

    result = differential_evolution(r_qual, bounds)
    sphere_r = result.x[0]
    data_sphere = sphere_form_factor(x_coord, sphere_r, 1)

    return x_coord[argrelextrema(data_sphere, np.less)[0]]


def get_symmetry_scores(image_data, center=None, num_components=50):
    """Compute vestor of symmetry scores by autocorrelation and fourier decomposition

    Keyword arguments:
    image_data -- 2D image array
    center -- tuple of center coordinates [default: center of image]
    num_components -- number of symmetry components in output

    Retuns:
    symmetry_scores -- 1D array of symmetry scores
    """

    if center is None:
        center = [(x-1)/2 for x in image_data.shape]

    edge_distance = min(image_data.shape[0] - center[0], center[0], \
                        image_data.shape[1] - center[1], center[1])

    polar_data = polar_projection(image_data, center, r_max=edge_distance, cval=0)
    polar_mask = polar_projection(image_data > 0, center, r_max=edge_distance, cval=1)

    psd_data = polar_data.sum(axis=1)
    psd_norm = polar_mask.sum(axis=1)
    psd_data[psd_norm > 0] /= psd_norm[psd_norm > 0]

    min_vals = list(get_psd_minima(psd_data))
    for val in [0, len(psd_data)]:
        if val not in min_vals:
            min_vals.append(val)
    fringe_edges = sorted(min_vals)
    fringe_ranges = list(zip(fringe_edges[:-1], fringe_edges[1:]))

    scores = np.zeros((len(fringe_ranges), num_components))

    for i, val in enumerate(fringe_ranges):
        start, end = val
        selected_data = polar_data[start:end]
        selected_data = selected_data[selected_data.min(axis=1) > 0]

        corr_data = np.zeros((selected_data.shape[0], selected_data.shape[1]+1))
        for j, polar_line in enumerate(selected_data):
            corr_data[j, :] = sp.correlate(polar_line, np.concatenate((polar_line, polar_line)))

        corr_t = np.mean(corr_data, axis=0)
        if corr_t[0] > 0:
            fringe_scores = dct(corr_t, type=1)[::2]
            scores[i] = fringe_scores[:num_components]
        else:
            scores[i] = 1

    return scores


def polar_projection(data, center, r_max=None, cval=None):
    """Compute polar projection

    Keyword arguments:
    data -- 2D image
    center -- tuple of center coordinates
    r_max -- maximum radius value in polar coordinates
    cval -- default value to fill missed areas

    Return:
    polar_data -- 2d image in polar coordinates with shape = (r_max, int(2*np.pi*r_max))
    """

    if r_max is None:
        x_max = max(data.shape[1] - center[1], center[1])
        y_max = max(data.shape[0] - center[0], center[0])
        r_max = np.sqrt(x_max**2 + y_max**2)

    r_max = int(r_max)

    r_i = np.arange(r_max)

    t_max = np.pi
    t_min = -np.pi
    t_num = int(2*np.pi*r_max)
    t_i = np.linspace(t_min, t_max, t_num, endpoint=False)

    t_grid, r_grid = np.meshgrid(t_i, r_i)

    x_grid = r_grid * np.cos(t_grid)
    y_grid = r_grid * np.sin(t_grid)

    x_grid, y_grid = x_grid.flatten(), y_grid.flatten()
    coords = np.vstack((y_grid + center[0], x_grid + center[1]))
    mapped_data = map_coordinates(data, coords, order=0, cval=cval)

    return mapped_data.reshape((r_max, t_num))
